function escapeKey(val) {
  return /^\w+$/.test(val) ? val : `"${val}"`;
}
function getType(val) {
  const type = typeof val;
  if (type === "undefined" || val === null) {
    return null;
  }
  if (Array.isArray(val)) {
    return "array";
  }
  return type;
}
function isObject(val) {
  return val !== null && !Array.isArray(val) && typeof val === "object";
}
function unique(arr) {
  return Array.from(new Set(arr));
}

function resolveSchema(obj, preserveResolve) {
  const schema = _resolveSchema(obj, obj, void 0, preserveResolve);
  return schema;
}
function _resolveSchema(input, parent, root, preserveResolve) {
  if (!isObject(input)) {
    return normalizeSchema({
      type: getType(input),
      default: input
    });
  }
  const node = {...input};
  const schema = {...node.$schema};
  let proxifiedNode = node;
  const getSchema = (key) => {
    schema.properties = schema.properties || {};
    if (!schema.properties[key]) {
      schema.properties[key] = _resolveSchema(node[key], proxifiedNode, root || proxifiedNode, preserveResolve);
    }
    return schema.properties[key];
  };
  proxifiedNode = new Proxy(node, {get: (_, key) => getSchema(key).default || node[key], set: () => false});
  for (const key in node) {
    if (key === "$resolve" || key === "$schema" || key === "$default") {
      continue;
    }
    getSchema(key);
  }
  if ("$default" in node) {
    schema.default = node.$default;
  }
  if (typeof node.$resolve === "function") {
    schema.default = node.$resolve(schema.default, parent, root || proxifiedNode);
    if (preserveResolve) {
      schema.resolve = node.$resolve;
    }
  }
  if (!schema.type) {
    schema.type = getType(schema.default) || (schema.properties ? "object" : "any");
  }
  return normalizeSchema(schema);
}
function normalizeSchema(schema) {
  if (schema.type === "array" && !("items" in schema)) {
    schema.items = {
      type: unique(schema.default.map((i) => getType(i)))
    };
    if (!schema.items.type.length) {
      schema.items.type = "any";
    }
  }
  return schema;
}

const TYPE_MAP = {
  array: "any[]",
  bigint: "bigint",
  boolean: "boolean",
  number: "number",
  object: "any",
  any: "any",
  string: "string",
  symbol: "Symbol",
  function: "Function"
};
const SCHEMA_KEYS = ["items", "default", "resolve", "properties", "title", "description", "$schema", "type"];
function generateTypes(schema, name = "MyObject") {
  return `interface ${name} {
  ` + _genTypes(schema, " ").join("\n ") + "\n}";
}
function _genTypes(schema, spaces) {
  const buff = [];
  for (const key in schema.properties) {
    const val = schema.properties[key];
    buff.push(...generateJSDoc(val));
    if (val.type === "object") {
      buff.push(`${escapeKey(key)}: {`, ..._genTypes(val, spaces + " "), "},\n");
    } else {
      let type;
      if (val.type === "array") {
        const _type = getTsType(val.items.type);
        type = _type.includes("|") ? `(${_type})[]` : `${_type}[]`;
      } else {
        type = getTsType(val.type);
      }
      buff.push(`${escapeKey(key)}: ${type},
`);
    }
  }
  if (buff.length) {
    const last = buff.pop() || "";
    buff.push(last.substr(0, last.length - 1));
  } else {
    buff.push("[key: string]: any");
  }
  return buff.map((i) => spaces + i);
}
function getTsType(type) {
  if (Array.isArray(type)) {
    return unique(type.map((t) => getTsType(t))).join(" | ") || "any";
  }
  return type && TYPE_MAP[type] || "any";
}
function generateJSDoc(schema) {
  let buff = [];
  if (schema.title) {
    buff.push(schema.title);
  }
  if (schema.description) {
    buff.push(schema.description);
  }
  if (schema.type !== "object" && schema.type !== "any" && !(Array.isArray(schema.default) && schema.default.length === 0)) {
    const stringified = JSON.stringify(schema.default);
    if (stringified) {
      buff.push(`@default ${stringified.replace(/\*\//g, "*\\/")}`);
    }
  }
  for (const key in schema) {
    if (!SCHEMA_KEYS.includes(key)) {
      buff.push("", `@${key} ${schema[key]}`);
    }
  }
  buff = buff.map((i) => i.split("\n")).flat();
  if (buff.length) {
    return buff.length === 1 ? ["/** " + buff[0] + " */"] : ["/**", ...buff.map((i) => ` * ${i}`), "*/"];
  }
  return [];
}

function applyDefaults(ref, input) {
  const schema = resolveSchema(ref, true);
  return _applyDefaults(schema, input);
}
function _applyDefaults(schema, input, root) {
  if (schema.type !== "object") {
    return {};
  }
  const merged = {};
  for (const key in schema.properties) {
    const prop = schema.properties[key];
    if (prop.type === "object") {
      merged[key] = _applyDefaults(prop, input == null ? void 0 : input[key], root || input);
    } else if (typeof prop.resolve === "function") {
      merged[key] = prop.resolve(input == null ? void 0 : input[key], merged, root || input);
    } else {
      merged[key] = (input == null ? void 0 : input[key]) || prop.default;
    }
  }
  return merged;
}

export { applyDefaults, generateTypes, resolveSchema };
